

"""
Evalulate COSMO-GHG clouds with DWD observations in Berlin

Berlin stations:
    - Berlin-Dahlem (id: 403)
    - Berlin-Schonefeld (id: 427)
    - Berlin-Tegel (id: 430)

data downloaded here:
    ftp://ftp-cdc.dwd.de/pub/CDC/observations_germany/climate/hourly/cloudiness/historical/
"""

from datetime import datetime, timedelta

ID2LOCATION = {
    403: (13.3017, 52.4537),
    427: (13.5306, 52.3807),
    430: (13.3088, 52.5644), # lon, lat
}


"""
Evalulate COSMO-GHG clouds with MODIS cloud mask
"""

import glob
import os

from scipy.interpolate import InterpolatedUnivariateSpline

import matplotlib.pyplot as plt
import netCDF4
import numpy as np
import pandas
import shapely
import scipy.interpolate
import scipy.ndimage as ndi
import xarray

import amrs.misc
import omi
import level2

from . import map
from . import misc
from . import config as cfg
from . import io


def plot_monthly_timeseries(ts, threshold, fig=None, marker='s',
                            ignore_mask=False):

    data = {}

    colors = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd',
              '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']

    if fig is None:
        fig = plt.figure(figsize=(5,3.5))

    for i,col in enumerate(ts.columns):
        print(col)

        if ignore_mask and 'mask' in col:
            continue

        mean = np.array([ts[col][ts.index.month == m].mean() for m in
                         range(1,13)])
        data[col] = mean

        if 'mask' in col:
            marker = 's'
            color = colors[3] if '1km' in col else colors[4]

        elif 'MODIS' in col:
            marker = 'v'
            color = colors[2]
        else:
            color = colors[0] if '1km' in col else colors[1]
            marker = 'o'

        plt.plot(np.arange(1,13), mean, label=col, marker=marker, color=color)


        #std = np.array([ts[col][ts.index.month == m].std() for m in
        #                 range(1,13)])
        #plt.errorbar(np.arange(1,13), mean, std, label=col, marker=marker,
        #             capsize=2)

    lines = sorted(plt.gca().get_lines(), key=lambda x: x.get_label())
    plt.legend(lines, [l.get_label() for l in lines])


    plt.xlim(1,12)
    plt.xticks(np.arange(1,13))

    plt.ylim(0,1)
    plt.yticks(np.arange(0,1.01,0.1))
    plt.ylabel('fraction of cloud-free pixels')
    plt.xlabel('month of 2015')
    plt.grid(True)

    return data


def cloud_free_timeseries(filenames, threshold=0.01, resolution='5km'):

    if resolution == '5km':
        columns = [
            'MODIS cloud mask (%s)' % resolution,
            'MODIS cloud fraction > %d%% (%s)' % (100*threshold, resolution),
            'COSMO cloud fraction > %d%% (%s)' % (100*threshold, resolution)
        ]
    else:
        columns = [
            'MODIS cloud mask (%s)' % resolution,
            'COSMO cloud fraction > %d%% (%s)' % (100*threshold, resolution)
        ]

    times = []
    data = []

    for filename in filenames:
        d = xarray.open_dataset(filename)

        modis_mask = d['Cloud_Mask_%s' % resolution].data

        if resolution == '5km':
            modis_fraction = d['Cloud_Fraction'].data

        cosmo_fraction = d['COSMO_Cloud_Fraction'].data
        time = pandas.Timestamp(d['COSMO_Cloud_Fraction'].time)

        d.close()

        valids = np.isfinite(modis_mask)

        if resolution == '5km':
            valids &= np.isfinite(modis_fraction)

        valids &= np.isfinite(cosmo_fraction)

        size = np.sum(valids)

        times.append(time)
        if resolution == '5km':
            data.append([
                1.0 - np.sum(modis_mask[valids]) / size,
                1.0 - np.sum(modis_fraction[valids] > threshold) / size,
                1.0 - np.sum(cosmo_fraction[valids] > threshold) / size
            ])
        else:
            data.append([
                1.0 - np.sum(modis_mask[valids]) / size,
                1.0 - np.sum(cosmo_fraction[valids] > threshold) / size
            ])

    return pandas.DataFrame(data, index=times, columns=columns)






def compute_pixel_boundaries(x0,y0):
    """\
    Compute boundaries of pixel using centre coords (x,y) by
    linear interpolation and extrapolation.
    """
    n,m = x0.shape

    x = np.zeros((n+1,m+1))
    y = np.zeros((n+1,m+1))

    # linear interpolation
    kernel = np.ones((2,2)) / 4.0
    x[1:-1,1:-1] = ndi.convolve(x0, kernel)[:-1,:-1]
    y[1:-1,1:-1] = ndi.convolve(y0, kernel)[:-1,:-1]

    # linerar extrapolation
    for i in range(1,n):
        f = InterpolatedUnivariateSpline(np.arange(1,m), x[i,1:-1], k=1)
        x[i,0] = f(0)
        x[i,m] = f(m)

        f = InterpolatedUnivariateSpline(np.arange(1,m), y[i,1:-1], k=1)
        y[i,0] = f(0)
        y[i,m] = f(m)


    for j in range(0,m+1):
        f = InterpolatedUnivariateSpline(np.arange(1,n), x[1:-1,j], k=1)
        x[0,j] = f(0)
        x[n,j] = f(n)

        f = InterpolatedUnivariateSpline(np.arange(1,n), y[1:-1,j], k=1)
        y[0,j] = f(0)
        y[n,j] = f(n)

    # 
    x = np.array([
        x[:-1,:-1],
        x[1:,:-1],
        x[1:,1:],
        x[:-1,1:],
    ])
    y = np.array([
        y[:-1,:-1],
        y[1:,:-1],
        y[1:,1:],
        y[:-1,1:],
    ])

    return x.swapaxes(0,1).swapaxes(1,2), y.swapaxes(0,1).swapaxes(1,2)


def read_cloud_mask(filename, varname):
    """
    Returns masked arrary where MODIS pixels are cloudy (00) or
    probably cloudy (00 and 01)

    MODIS cloud mask is a bit array with bit fields 1 and 2 decode
    the FOV quality flag with:

    - 00: cloudy
    - 01: uncertain clear (probably cloudy)
    - 10: probably clear
    - 11: clear
    """
    with netCDF4.Dataset(filename) as nc:

        if varname == 'Cloud_Mask':
            mask = nc.variables[varname][0].data
        else:
            mask = nc.variables[varname][:,:,0].data


    # check if 1st and 2nd bits are 1 and 1
    first = (mask & 2**2 != 0)

    return first == False # cloudy and probably cloudy

    second = (mask & 2**1 != 0)
    return (first == False) & (second == False) # cloudy


def get_lonlat(filename, varname):
    """
    Returns longitude/latitude at cloud mask at 1km resolution.

    The MODIS file stores only 5km lon/lat which needs to be interpolated to
    the 1km resolution.
    """
    with netCDF4.Dataset(filename) as nc:

        time = nc.variables['Scan_Start_Time'][:].data

        lon = nc.variables['Longitude'][:].data
        lat = nc.variables['Latitude'][:].data

        if varname == 'Cloud_Mask':
            nx, ny = nc.variables['Cloud_Mask'].shape[1:]
        else:
            nx, ny = nc.variables[varname].shape[:2]

    if (nx, ny) != lon.shape:

        # lon/lat are center coords on a 5x5 km grid
        x = np.linspace(0,1,nx)
        y = np.linspace(0,1,ny)

        X, Y = np.meshgrid(x,y)

        # coords of lon and lat at low resolution
        xg = x[2::5][:lon.shape[0]]
        yg = y[2::5][:lat.shape[1]]

        # interpolate time and lon/lat to high resolution
        ifunc = scipy.interpolate.RectBivariateSpline(xg,yg,time)
        time = ifunc(x,y)

        ifunc = scipy.interpolate.RectBivariateSpline(xg,yg,lon)
        lon = ifunc(x,y)

        ifunc = scipy.interpolate.RectBivariateSpline(xg,yg,lat)
        lat = ifunc(x,y)

    # time (seconds after 1993-01-01)
    offset = np.datetime64('1993-01-01', 'ns') - np.datetime64('1970-01-01', 'ns')
    time = np.array((1e9 * time).astype('int'), dtype='datetime64[ns]') + offset

    return time, lon, lat



def read_modis(filename, resolution='5km'):

    varnames = {
        '1km': ['Cloud_Mask_1km'],
        '5km': ['Cloud_Mask_5km', 'Cloud_Fraction']
    }[resolution]

    data = xarray.Dataset()
    dims = ('along_%s' % resolution, 'across_%s' % resolution)
    cdims = dims + ('corners', )

    for varname in varnames:
        if varname.startswith('Cloud_Mask'):
            values = read_cloud_mask(filename, varname=varname)
        else:
            with netCDF4.Dataset(filename) as nc:
                values = nc.variables[varname][:].data

        data[varname] = xarray.DataArray(values, dims=dims)

    # aux data
    time, lon, lat = get_lonlat(filename, varname=varname)
    lonc, latc = compute_pixel_boundaries(lon, lat)

    # convert to rotated pole coords
    rlon, rlat = misc.rotpole2wgs(lon, lat, inverse=True)

    # check if any observation in model domain
    points = shapely.geometry.MultiPoint(np.transpose([rlon.flatten(),
                                                       rlat.flatten()]))
    if not cfg.DOMAIN.intersects(points):
        return None

    # compute pixel area
    deg2_to_km2 = (111.32 * 40075.0 * np.cos(np.deg2rad(lat)) / 360.0)
    area = np.zeros(lon.shape)

    for i,j in np.ndindex(area.shape):
        px = np.transpose([lonc[i,j,:], latc[i,j,:]])
        area[i,j] = shapely.geometry.Polygon(px).area * deg2_to_km2[i,j]

    data['time'] = xarray.DataArray(time, dims=dims)
    data['longitude'] = xarray.DataArray(lon, dims=dims)
    data['latitude'] = xarray.DataArray(lat, dims=dims)

    data['longitude_corners'] = xarray.DataArray(lonc, dims=cdims)
    data['latitude_corners'] = xarray.DataArray(latc, dims=cdims)

    data['Pixel_Area'] = xarray.DataArray(area, dims=dims)

    return data



def make_level3(lonc, latc, values, method='cvm', grid=None):
    """
    Make Level3 map.
    """

    # corners of COSMO grid
    rlonc, rlatc = cfg.rlonc, cfg.rlatc

    # 
    valids = np.isfinite(values)

    if method == 'hist':

        # bin to COSMO grid
        H, _, _ = np.histogram2d(rlat[valids], rlon[valids], bins=[rlatc, rlonc],
                                 weights=values[valids])

        N, _, _ = np.histogram2d(rlat[valids], rlon[valids], bins=[rlatc, rlonc])

        values = H / N

    elif method == 'cvm':


        # values
        if grid is None:
            grid = omi.create_grid(cfg.startlon, cfg.startlat, cfg.stoplon,
                                   cfg.stoplat, cfg.dlon, cfg.dlat, cfg.pollon,
                                   cfg.pollat)

        # swap axes for gridding
        lonc = lonc.swapaxes(0,2)
        latc = latc.swapaxes(0,2)

        omi.cvm_grid(grid, lonc, latc, values, np.zeros(values.shape),
                     np.ones(values.shape), ~valids)

        values = (grid['values'] / grid['weights']).T

    else:
        raise ValueError


    # fill small gaps
    #is_cloudy = amrs.misc.normalized_convolution(is_cloudy, np.ones((3,3)))
    #is_cloudy = is_cloudy > 0.5

    #if varname.startswith('Cloud_Mask'):
    #    dtype = 'bool'
    #else:
    dtype = 'f4'

    return np.ma.array(values, dtype=dtype, mask=np.isnan(values))



def cloud_fraction(is_cloudy):

    if hasattr(is_cloudy, 'mask'):
        valids = ~is_cloudy.mask
    else:
        valids = Ellipsis

    cloudy = np.sum(is_cloudy[valids] == True)
    clear = np.sum(is_cloudy[valids] == False)

    fraction = clear / (cloudy + clear)

    return clear, cloudy, fraction



def mask_inside_domain(lon, lat):
    rlon, rlat = misc.rotpole2wgs(np.array(lon), np.array(lat),
                                  inverse=True)

    mask = (cfg.startlon-0.1 <= rlon) & (rlon <= cfg.stoplon+0.1)
    mask &= (cfg.startlat-0.1 <= rlat) & (rlat <= cfg.stoplat+0.1)

    return mask


def clip_on_cosmo(lon, lat, resolution='5km'):

    mask = mask_inside_domain(lon, lat)

    # make rectangular mask
    i = np.arange(mask.shape[0])[mask.any(axis=1)]
    j = np.arange(mask.shape[1])[mask.any(axis=0)]

    return {'across_%s' % resolution: j, 'along_%s' % resolution: i}



def iter_orbits(filenames, resolution='5km'):

    data = None
    values = None

    for i, filename in enumerate(filenames):

        # time from filename
        dd, tt = os.path.basename(filename).split('.')[1:3]
        mean_time = datetime.strptime(dd+tt, 'A%Y%j%H%M')

        if i == 0:
            last_time = mean_time

        if data is not None and (mean_time - last_time) > timedelta(minutes=60):
            slicing = clip_on_cosmo(data.longitude, data.latitude, resolution)
            yield data[slicing]

            data = None

        #print(i,filename)
        values = read_modis(filename, resolution)
        last_time = mean_time

        if values is not None:
            if data is None:
                data = values
            else:
                data = xarray.concat([data, values], dim='along_%s' %
                                     resolution)

    if values is not None:
        slicing = clip_on_cosmo(data.longitude, data.latitude,
                                resolution=resolution)
        yield data[slicing]



def grid_orbit(lonc, latc, values):

    grid = omi.create_grid(cfg.startlon, cfg.startlat, cfg.stoplon,
                           cfg.stoplat, cfg.dlon, cfg.dlat, cfg.pollon,
                           cfg.pollat)

    #if value.mask.all(): # all cells are masked
    #    last_time = mean_time
    #    continue

    #if not (6 <= mean_time.round('H').hour <= 18): # only daytime
    #    print('Reject time: %s' % mean_time)
    #    last_time = mean_time
    #    continue

    # concat lists

    modis_mask = np.ma.array(modis_mask)

    if reduce:
        modis_mask = modis_mask.mean(axis=0) > 0.5

        mean_time = 0.0
        for i,t in enumerate(times):
            t = (t - datetime(1993,1,1)).total_seconds()
            mean_time += t * valids[i] / sum(valids)

        mean_time = datetime(1993,1,1) + timedelta(seconds=mean_time)
        mean_time = pandas.Timestamp(mean_time)

        return mean_time, modis_mask

    return np.array(times), modis_mask


def average_datetime(time):
    """
    Average a dateime64 array.
    """
    ref = np.datetime64('2015-01-01', 'ns')
    delta = np.mean((time - ref).astype(int))
    delta = int(delta)

    return ref + np.timedelta64(delta, 'ns')



def plot_cloud_fraction(data, resolution='5km', output_path='.', close=True):

    mean_time = pandas.Timestamp(data.COSMO_Cloud_Fraction.time)

    # MODIS cloud fraction
    if resolution == '5km':
        fig = map.make_level2_map(data.longitude_corners,
                                  data.latitude_corners,
                                  data['Cloud_Fraction'],
                                  vmin=0.0, vmax=1.0, cmap='Blues_r', xlim=None,
                                  ylim=None, label='MODIS cloud fraction',
                                  fig_size=5)
    else:
        fig = map.make_level2_map(data.longitude_corners,
                                  data.latitude_corners,
                                  data['Cloud_Mask_1km'],
                                  vmin=0.0, vmax=1.0, cmap='Blues_r', xlim=None,
                                  ylim=None, cb_labels=['clear', 'cloudy'],
                                  fig_size=5)

    fig.suptitle(mean_time.strftime('%Y-%m-%d %H:%M UTC'),
                 bbox=dict(facecolor='white'), fontsize=12)

    fig_filename = os.path.join(
        output_path, 'figs',
        mean_time.strftime('%%Y%%m%%d.%%H%%M_MODIS_Cloud_Fraction_%s.png' %
                           resolution)
    )
    fig.savefig(fig_filename)

    if close:
        plt.close()


    # COSMO cloud fraction
    fig = map.make_level2_map(data.longitude_corners,
                              data.latitude_corners,
                              data['COSMO_Cloud_Fraction'],
                              vmin=0.0, vmax=1.0, cmap='Blues_r', xlim=None,
                              ylim=None, label='COSMO cloud fraction',
                              fig_size=5)

    fig.suptitle(mean_time.round('H').strftime('%Y-%m-%d %H:%M UTC'),
                 bbox=dict(facecolor='white'), fontsize=12)


    fig_filename = os.path.join(
        output_path, 'figs',
        mean_time.strftime('%%Y%%m%%d.%%H%%M_COSMO_Cloud_Fraction_%s.png' %
                           resolution)
    )
    fig.savefig(fig_filename)

    if close:
        plt.close()




def main(day=None, resolution='5km', satellite='Terra'):

    day = '???' if day is None else '%03d' % day

    if satellite == 'Terra':
        filenames = sorted(glob.glob("/input/MODIS/MOD06_L2/2015/%s/*.hdf" % day))
    elif satellite == 'Aqua':
        filenames = sorted(glob.glob("/input/MODIS/MYD06_L2/2015/%s/*.hdf" % day))
    else:
        raise ValueError

    data = None

    for data in iter_orbits(filenames, resolution):

        mean_time = pandas.Timestamp(average_datetime(data.time))

        if mean_time > datetime(2015,12,25):
            continue # no COSMO simulation

        starttime = {
            True: datetime(2015,7,1),
            False: datetime(2015,1,1)
        }[mean_time.round('H') >= datetime(2015,7,1)]

        # read COSMO cloud mask
        clct = io.read_field(mean_time.round('H'), starttime, keys=['CLCT'],
                             input_files='2D',
                             path='/output/SMARTCARB/output/')[0]

        # clct -> Level2
        mask = mask_inside_domain(data.longitude, data.latitude)
        clct = level2.sum_satellite_pixels([clct], data.longitude_corners.data,
                                           data.latitude_corners.data,
                                           ignore_pixels=~mask)[0]

        dims = ('along_%s' % resolution, 'across_%s' % resolution)
        data['COSMO_Cloud_Fraction'] = xarray.DataArray(clct, dims=dims,
                                                        attrs={'time': str(mean_time)})

        output_path = '/output/SMARTCARB/output/Clouds/MODIS-%s' % satellite
        if satellite == 'Terra':
            pattern = 'MOD06_L2.%%Y%%m%%d.%%H%%M_%s.nc' % resolution
        elif satellite == 'Aqua':
            pattern = 'MYD06_L2.%%Y%%m%%d.%%H%%M_%s.nc' % resolution

        output_filename = os.path.join(
            output_path,
            mean_time.strftime('%Y-%m'),
            mean_time.strftime(pattern)
        )

        os.makedirs(os.path.dirname(output_filename), exist_ok=True)

        print('->', output_filename)
        data.to_netcdf(output_filename)

        # figures
        plot_cloud_fraction(data, resolution, output_path)

    return data


